@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010-2017 Phusion Holding B.V.
+ *  Copyright (c) 2010-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -60,12 +60,13 @@
 #include "DirectoryMapper.h"
 #include "Utils.h"
 #include <modp_b64.h>
+#include <WrapperRegistry/Registry.h>
 #include <FileTools/FileManip.h>
 #include <FileTools/FileManip.h>
 #include <Utils.h>
-#include <Utils/IOUtils.h>
-#include <Utils/StrIntUtils.h>
-#include <Utils/SystemTime.h>
+#include <IOTools/IOUtils.h>
+#include <StrIntTools/StrIntUtils.h>
+#include <SystemTools/SystemTime.h>
 #include <Utils/HttpConstants.h>
 #include <Utils/ReleaseableScopedPointer.h>
 #include <LoggingKit/LoggingKit.h>
@@ -232,9 +233,11 @@ private:
 	enum Threeway { YES, NO, UNKNOWN };
 
 	Threeway m_hasModRewrite, m_hasModDir, m_hasModAutoIndex, m_hasModXsendfile;
+	WrapperRegistry::Registry wrapperRegistry;
 	CachedFileStat cstat;
 	WatchdogLauncher watchdogLauncher;
 	boost::mutex cstatMutex;
+	boost::mutex configMutex;
 
 	static Json::Value strsetToJson(const set<string> &input) {
 		Json::Value result(Json::arrayValue);
@@ -443,9 +446,12 @@ private:
 	bool prepareRequest(request_rec *r, DirConfig *config, const char *filename, bool coreModuleWillBeRun = false) {
 		TRACE_POINT();
 
-		DirectoryMapper mapper(r, config, &cstat, &cstatMutex, serverConfig.statThrottleRate);
+		DirectoryMapper mapper(r, config, wrapperRegistry, &cstat,
+		    &cstatMutex, serverConfig.statThrottleRate, &configMutex);
 		try {
-			if (mapper.getApplicationType() == PAT_NONE) {
+			if (config->getAppStartCommand().empty()
+			 && mapper.getDetectorResult().isNull())
+			{
 				// (B) is not true.
 				disableRequestNote(r);
 				return false;
@@ -983,8 +989,18 @@ private:
 		}
 
 		// Phusion Passenger options.
-		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_ROOT"), mapper.getAppRoot());
-		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_TYPE"), mapper.getApplicationTypeName());
+		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_ROOT"),
+			mapper.getAppRoot());
+		if (!config->getAppStartCommand().empty()) {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_START_COMMAND"),
+				config->getAppStartCommand());
+		} else if (mapper.getDetectorResult().wrapperRegistryEntry != NULL) {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_TYPE"),
+				mapper.getDetectorResult().wrapperRegistryEntry->language);
+		} else {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_START_COMMAND"),
+				mapper.getDetectorResult().appStartCommand);
+		}
 		constructRequestHeaders_autoGenerated(r, config, result);
 
 		/*********************/
@@ -1247,6 +1263,7 @@ public:
 	    : cstat(1024),
 	      watchdogLauncher(IM_APACHE)
 	{
+		wrapperRegistry.finalize();
 		postprocessConfig(s, pconf, ptemp);
 
 		Json::Value loggingConfig;
@@ -1319,6 +1336,8 @@ public:
 		config["instance_registry_dir"] = nonEmptyString(serverConfig.instanceRegistryDir);
 		config["security_update_checker_disabled"] = serverConfig.disableSecurityUpdateCheck;
 		config["security_update_checker_proxy_url"] = nonEmptyString(serverConfig.securityUpdateCheckProxy);
+		config["telemetry_collector_disabled"] = serverConfig.disableAnonymousTelemetry;
+		config["telemetry_collector_proxy_url"] = nonEmptyString(serverConfig.anonymousTelemetryProxy);
 		config["user_switching"] = serverConfig.userSwitching;
 		config["default_user"] = serverConfig.defaultUser.toString();
 		config["default_group"] = serverConfig.defaultGroup.toString();
@@ -1326,10 +1345,15 @@ public:
 		config["show_version_in_header"] = serverConfig.showVersionInHeader;
 		config["max_pool_size"] = serverConfig.maxPoolSize;
 		config["pool_idle_time"] = serverConfig.poolIdleTime;
+		config["max_instances_per_app"] = serverConfig.maxInstancesPerApp;
 		config["response_buffer_high_watermark"] = serverConfig.responseBufferHighWatermark;
 		config["stat_throttle_rate"] = serverConfig.statThrottleRate;
 		config["turbocaching"] = serverConfig.turbocaching;
 		config["prestart_urls"] = strsetToJson(serverConfig.prestartURLs);
+		config["admin_panel_url"] = nonEmptyString(serverConfig.adminPanelUrl);
+		config["admin_panel_auth_type"] = nonEmptyString(serverConfig.adminPanelAuthType);
+		config["admin_panel_username"] = nonEmptyString(serverConfig.adminPanelUsername);
+		config["admin_panel_password"] = nonEmptyString(serverConfig.adminPanelPassword);
 
 		if (!serverConfig.logFile.empty()) {
 			config["log_target"] = serverConfig.logFile.toString();
